import {
	Injectable,
	InternalServerErrorException,
	NotFoundException,
	StreamableFile,
} from "@nestjs/common";
import { DbService } from "apps/backend/src/app/db/db.service";
import {
	FilesForMachinesTable,
	FilesGroupTable,
	FilesTable,
	FilesTypeForMachine,
	MachinesTable,
} from "apps/backend/src/app/db/schema";
import { StorageService } from "apps/backend/src/app/storage/storage.service";
import { data } from "autoprefixer";
import { eq, ilike } from "drizzle-orm";

@Injectable()
export class FilesService {
	constructor(
		private readonly storage: StorageService,
		private readonly database: DbService,
	) {}

	/**
	 * Retrieves a file and its related information from the database and associated storage.
	 *
	 * @param fileId The unique identifier for the file to be retrieved.
	 * @return A promise that resolves to an object containing the streamable file and additional file information.
	 * @throws NotFoundException if the file does not exist in the database.
	 */
	public async get(fileId: string): Promise<StreamableFile> {
		const foundFiles = await this.database
			.use()
			.select()
			.from(FilesTable)
			.where(eq(FilesTable.uuid, fileId))
			.prepare("getFileFromId")
			.execute();

		if (foundFiles.length === 0)
			throw new NotFoundException("File not found", {
				description: `Identifier : ${fileId}`,
			});

		if (foundFiles.length > 1)
			console.log(
				"Multiples entries found in the database.\nInformation from the first row will be used.",
			);
		const file = foundFiles[0];

		const fileBuffer = await this.storage.read(
			file.checksum,
			file.extension,
			file.isDocumentation,
		);
		const fileInformation = await this.storage.generateInformation(
			fileBuffer,
			file.fileName,
			file.isDocumentation,
		);
		const fileNameWithoutSpaces = file.fileName.replace(/\s/g, "_");
		return new StreamableFile(fileBuffer, {
			disposition: `attachment; filename="${fileNameWithoutSpaces}.${fileInformation.fileType[0].extension.toLowerCase()}"`,
		});
	}

	public async search(limit: number, offset: number, searchField: string) {
		return await this.database
			.use()
			.select()
			.from(FilesTable)
			.where(ilike(FilesTable.fileName, String(searchField)))
			.limit(limit)
			.offset(offset)
			.prepare("searchFiles")
			.execute();
	}

	//TODO save a file
	public async save(file: Buffer, data: Map<string, unknown>) {
		const _machineIds = Array.from(data.get("machineId") as string[]);
		const machinesIds = new Set<string>();
		console.log(
			`Checking if machine with ID ${_machineIds} exist in the database...`,
		);
		for (const machineId of _machineIds) {
			const machineExists = await this.database
				.use()
				.select({
					uuid: MachinesTable.id,
				})
				.from(MachinesTable)
				.where(eq(MachinesTable.id, machineId))
				.prepare("checkMachineExists")
				.execute();

			if (machineExists.length === 0) {
				throw new NotFoundException(`Machine with ID "${machineId}" not found`);
			}

			machinesIds.add(machineExists[0].uuid);
		}

		const _group = data.get("groupId") as string;
		// verify that the group exist in the database
		const groupExists = await this.database
			.use()
			.select()
			.from(FilesGroupTable)
			.where(eq(FilesGroupTable.uuid, _group))
			.prepare("checkGroupExists")
			.execute();

		if (groupExists.length === 0) {
			throw new NotFoundException(`Group with ID "${_group}" not found`);
		}

		try {
			const saveResult = await this.storage.new(
				data.get("fileName") as string,
				file,
				_machineIds,
				Boolean(data.get("isDocumentation")),
			);
			const inserted = await this.database
				.use()
				.insert(FilesTable)
				.values({
					fileName: data.get("fileName") as string,
					checksum: saveResult.fileChecksum,
					extension: saveResult.fileType[0].extension,
					fileSize: saveResult.fileSize,
					fileType: saveResult.fileType.mime,
					isRestricted: Boolean(data.get("isRestricted")),
					isDocumentation: Boolean(data.get("isDocumentation")),
					uploadedBy: data.get("uploadedBy") as string,
				})
				.returning();

			for (const machineId of machinesIds) {
				//TODO insert a link betwen fileId and MachineIds[]
				const linkRow = await this.database
					.use()
					.insert(FilesForMachinesTable)
					.values({
						fileId: inserted[0].uuid,
						machineId: machineId,
					});
			}
		} catch (e) {
			throw new InternalServerErrorException(
				"It seems that the insertion in the database failed.",
				{
					cause:
						process.env.NODE_ENV === "production"
							? "Internal server error"
							: {
									message: e.message,
									stack: e.stack,
								},
					description: `Nom de fichier : "${data.get("fileName")}" `,
				},
			);
		}
	}
}
